from .menu import Menu
from .auth import auth
import pyke
import click

class LumaCommand(click.Command):
  """
  Currently not used. Experimental override of defualt click command to add auth
  """
  def __init__(self, *args, **kwargs):
    ctx = self.make_context(kwargs.get('name'), list(args))
    do_auth = kwargs.get('auth')
    if do_auth is None or do_auth is True:
      profile_option = click.Option(['--profile', '-p'], help='The name of the profile to use as the context for the command.', autocompletion=auth.get_profile_names)
      ctx.params.update(profile_option)
    if do_auth is not None:
      del kwargs['auth']

    if 'profile' in ctx.params.keys():
      profile_name = ctx.params['profile']
      if not profile_name:
        profiles = auth.list_profiles()
        profile_data = Menu().get_user_select(profiles, format, sort='profileName')
        profile_name = profile_data.get('profileName')
      else:
        profile_data = auth.get_profile_data(profile_name)

      env_name = profile_data.get('env')
      env_data = Menu().sanatize_data(auth.get_env_data(env_name))

      ctx.obj = {**profile_data, **env_data}
      ctx.invoke(self.callback, *args, **kwargs)

def select_profile(ctx, param, value):
  profiles = auth.list_profiles()
  profile_names = [x.get('profileName') for x in profiles]
  if not value:
    format = "{Option} {profileName} {env} {orgName} {orgId}"
    titles = {"profileName": "Profile Name", "env": "Env", "orgName": "Org Name", "orgId": "Org ID"}
    profile_data = Menu().get_user_select(profiles, format, sort='profileName', titles=titles)
  else:
    if value not in profile_names:
      raise click.BadParameter(click.style('You must provide the name of a profile that you have configured', fg='red'))

    # Use this to update old profiles
    profile_data = auth.get_profile_data(value)
    if 'profileName' not in profile_data.keys():
      profile_data['profileName'] = value
      auth.update_profile(value, profile_data)

    if profile_data.get('orgId') is None or profile_data.get('orgId') == '':
      raise click.BadParameter(click.style('You must finish configuring this profile before you can use it', fg='red'))

  env_name = profile_data.get('env')
  env_data = Menu().sanatize_data(auth.get_env_data(env_name))

  ctx.obj = {**profile_data, **env_data}

  return pyke.Util(ctx=ctx.obj)



"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

import requests
import urllib3
import click
import pyke
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

class ExperienceBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')
    self.json           = kwargs.get('json', False)
    self.name           = kwargs.get('name')
    self.format         = kwargs.get('format')
    self.filter         = kwargs.get('filter')
    self.page           = kwargs.get('page')
    self.pagesize       = kwargs.get('pagesize')
    self.device_type    = kwargs.get('device_type')
    self.collection     = kwargs.get('collection')

    self.import_file = None
    # Expand the filename and get the absolute path. Relative paths won't resolve if entered via prompt
    filename = kwargs.get('import_file')
    if filename:
      try:
        self.import_file = os.path.abspath(os.path.expanduser(filename))
      except:
        raise pyke.CliException('Invalid value for --import-file')

    self.export_file = None
    filename = kwargs.get('export_file')
    if filename:
      try:
        self.export_file = os.path.abspath(os.path.expanduser(filename))
      except:
        raise pyke.CliException('Invalid value for --export-file')

    self.type = kwargs.get('type')
    self.redirect_url = kwargs.get('redirect_url')
    self.description = kwargs.get('description')
    self.label = kwargs.get('label')
    self.activation_code = kwargs.get('activation_code')
    self.is_template = kwargs.get('template', False)

    if self.type is None:
      self.type = 'template' if self.is_template else 'standard'

  def list(self):
    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/experiences?page={page}&pagesize={pagesize}&{filter}', {**data} ))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table(resp.get('payload', {}).get('data', {}), self.format)
    self.util.print_record_count(resp)

  def export(self):
    if not self.json and not self.export_file:
      try:
        self.export_file = input('Export file: ')
        if self.export_file is None or self.export_file == '':
          raise pyke.CliException('Must provide --export-file')

        self.export_file = os.path.abspath(os.path.expanduser(self.export_file))
      except:
        raise pyke.CliException('Invalid value for --export-file')


    if self.name and self.label:
      raise pyke.CliException('Must provide either --label or --name; not both')

    # Determine if we have to lookup the namespace
    if not self.name:
      if not self.label:
        self.label = input('Label: ')
        if self.label is None or self.label == '':
          raise pyke.CliException('Must provide --name or --label of the experience to export')

      experience_resp = self.util.cli_request('GET',
        self.util.build_url('{experienceCloudUri}/iot/v1/experiences?{filter}', {'filter': 'label={}'.format(self.label)} ))

      experience_data = experience_resp.get('payload', {}).get('data')
      if len(experience_data) == 0:
        raise pyke.CliException('Experience label "{}" not found'.format(self.label))

      if len(experience_data) > 1:
        raise pyke.CliException('More than one experience was found with the label "{}"'.format(self.label))

      self.name = experience_data[0].get('name')

    export_resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/experiences/{name}/export', {'name': self.name}))
    if self.json:
      click.echo(json.dumps(export_resp))
      return

    status_id = export_resp.get('payload', {}).get('data', {}).get('statusId', {})
    status_resp = self.util.show_progress(status_id, label='Exporting experience')

    error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
    if error_msg is not None:
      raise pyke.CliException('Export failed. {}'.format(error_msg))

    click.echo(' ')

    file_url = status_resp.get('payload', {}).get('data', {}).get('summary', {}).get('fileInfo', {}).get('signedUrl')
    file_resp = requests.get(file_url, verify=verify_tls)
    with open(self.export_file, 'wb') as f:
      f.write(file_resp.content)

    click.echo('Saved to {}'.format(self.export_file))

  def import_(self):
    if not self.util.ctx.get('instanceType'):
      raise click.ClickException(click.style("This profile looks old. Run 'luma profile refresh-token' to migrate this profile to the new format.", fg='red'))
    if self.util.ctx.get('instanceType') == 'cc':
      raise click.ClickException(click.style("This command must be run with a studio profile", fg='red'))

    if self.collection:
      collection = self.util.get_collection(self.collection)
    else:
      titles =  titles={'name': 'Name', 'id': 'Collection ID'}
      collection = pyke.Menu().get_user_select(self.util.get_all_collections(),'{name} {id}', titles=titles)

    collection_id = collection.get('id')

    if not self.json:
      click.echo('Uploading file...')

    # Request ephemeral token
    token_resp = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/files/ephemeral?contentType=application/json'))
    token_data = token_resp.get('payload', {}).get('data', {})
    token_fields = token_data.get('fields')

    aws_access_key_id = token_fields.get('AWSAccessKeyId')
    ephemeral_key = token_fields.get('key').replace('${filename}', 'import.json')
    policy = token_fields.get('policy')
    signature = token_fields.get('signature')
    url = token_data.get('url')

    # Post to ephemral URL
    data = {
      'AWSAccessKeyId': aws_access_key_id,
      'key': ephemeral_key,
      'policy': policy,
      'signature': signature,
      'Content-Type': 'application/json'
    }

    aws_resp = None
    with open(self.import_file, 'rb') as f:
      aws_resp = requests.post(url, data=data, files={'file': f}, verify=verify_tls)

    if not self.json:
      click.echo('File uploaded.')
      click.echo('')

    # Post to import
    import_json = {
      'name':'',
      'description': self.description,
      'type': self.type,
      'experienceCollectionId': collection_id,
      'label': self.label,
      'redirectUrl': self.redirect_url,
      'code': self.activation_code,
      'ephemeralKey': ephemeral_key,
      'device': self.device_type
    }

    import_resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/experiences/import'), json=import_json)
    if self.json:
      click.echo(json.dumps(import_resp))
      return

    status_id = import_resp.get('payload', {}).get('data', {}).get('statusId', {})
    status_resp = self.util.show_progress(status_id, label='Importing experience')

    error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
    if error_msg is not None:
      raise pyke.CliException('Import failed. {}'.format(error_msg))

    click.echo('Successfully imported experience.')

  def create(self):
    if not self.util.ctx.get('instanceType'):
      raise click.ClickException(click.style("This profile looks old. Run 'luma profile refresh-token' to migrate this profile to the new format.", fg='red'))
    if self.util.ctx.get('instanceType') == 'cc':
      raise click.ClickException(click.style("This command must be run with a studio profile", fg='red'))

    xp_data = {}
    xp_data['label'] = self.name
    xp_data['code'] = self.activation_code
    xp_data['description'] = self.description
    xp_data['device'] = self.device_type
    xp_data['type'] = self.type
    if self.redirect_url:
      xp_data['redirectUrl'] = self.redirect_url

    if self.collection:
      collection = self.util.get_collection(self.collection)
      xp_data['experienceCollectionId'] = collection.get('id')
    else:
      titles =  titles={'name': 'Name', 'id': 'Collection ID'}
      collection = pyke.Menu().get_user_select(self.util.get_all_collections(),'{name} {id}', titles=titles)
      xp_data['experienceCollectionId'] = collection.get('id')

    resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/experiences'), json=xp_data)
    if self.json:
      click.echo(resp)
      return

    titles =  titles={'label': 'Label', 'name': 'Namespace', 'code': 'Code', 'device': 'Device'}
    pyke.Menu().print_menu([resp['payload']['data']], '{label} {name} {code} {device}', titles=titles)



"""
    These sub-commands and their corresponding logic are both implamented in this file.
    TO DO: Break out the logic for these commands into a seperate behavior file.
"""

from pathlib import Path
import requests
import urllib3
import click
import pyke
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

@click.group('org', help='Get information about Orgs such as parent-child org relationships')
def organization():
  pass

@click.command('ls', help="List the Orgs you can make profiles for. \
The 'isTest' column in the returned table indicates if a studio is a test org or a prod org.")
@click.option('--env', help='The Env to point to')
@click.option('--format', '-f', default='{id} {name} {instanceType} {isTest}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json from the platform.')
def list(env, format, filter, json_flag):
  handlers = {
    'instanceType': lambda x: 'dev' if x == 'cc' else x
  }

  if 'instanceType=dev' in filter:
    filter = filter.replace('instanceType=dev', 'instanceType=cc')

  if 'isTest=None' in filter:
    filter = filter.replace('isTest=None', 'instanceType=cc')
  
  env_format = '{envName} {app}'
  titles = {'envName': 'Env Name', 'app': 'App'}

  if not env:
    envs = pyke.auth.list_envs()
    env_data = pyke.Menu().get_user_select(envs, env_format, titles=titles)
  else:
    env_data = pyke.auth.get_env_data(env)

  util = pyke.Util(ctx=env_data)

  access_token = util.login_without_context()
  util.ctx.update({'accessToken': access_token})

  headers = {
      'Authorization': 'Bearer ' + access_token,
      'Content-Type': 'application/json'
  }

  app_name = env_data.get('app')
  if not app_name:
    raise click.ClickException(click.style("Env not configured", fg='red'))

  resp = requests.get('{}/auth/v1/me/companies?sort=name&{}'.format(app_name, filter), headers=headers, verify=verify_tls)

  if json_flag:
      click.echo(resp.text)
      return

  pyke.Menu().print_menu(resp.json()['payload']['data'], format, handlers=handlers)

@click.command('child-orgs', help='List the Orgs that a parent Org can share objects with')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {instanceType} {isTest}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--filter', default='')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json from the platform.')
def child_companies(profile, format, filter, json_flag):
  util = pyke.Util()

  if 'instanceType=dev' in filter:
    filter = filter.replace('instanceType=dev', 'instanceType=cc')

  if 'isTest=None' in filter:
      filter = filter.replace('isTest=None', 'instanceType=cc')

  handlers = {
    'instanceType': lambda x: 'dev' if x == 'cc' else x
  }

  resp = util.cli_request('GET',
    util.build_url('{app}/auth/v1/company/child-companies?{filter}', {'filter': filter}))

  if json_flag:
    click.echo(json.dumps(resp))
    return

  util.print_table(resp['payload']['data'], format, handlers=handlers)

organization.add_command(list)
organization.add_command(child_companies)



"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

import click
import pyke
import json

class ExperienceCollectionBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')

    self.json           = kwargs.get('json')
    self.format         = kwargs.get('format')
    self.filter         = kwargs.get('filter')
    self.page           = kwargs.get('page')
    self.pagesize       = kwargs.get('pagesize')
    self.name           = kwargs.get('name')

  def list(self):
    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}
    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/experience-collections?page={page}&pagesize={pagesize}&{filter}', {**data} ))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table(resp.get('payload', {}).get('data', {}), self.format)
    self.util.print_record_count(resp)





"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

from pathlib import Path
import requests
import click
import pyke
import time
import json
import sys
import os

class ComponentSetVersionBehavior:
  def __init__(self, **kwargs):
    self.util            = kwargs.get('profile')
    self.version         = kwargs.get('version')
    self.component_set   = kwargs.get('component_set')
    self.json            = kwargs.get('json')
    self.filter          = kwargs.get('filter')
    self.page            = kwargs.get('page')
    self.pagesize        = kwargs.get('pagesize')
    self.format          = kwargs.get('format')
    self.label           = kwargs.get('label')
    self.from_version    = kwargs.get('from_version')
    self.version_bump    = kwargs.get('version_bump')
    self.css_includes    = kwargs.get('css_includes')
    self.direct_includes = kwargs.get('direct_includes')

    self.component_set_file_path  = kwargs.get('component_set_file_path')

  def list(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')

    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

    titles = {
          'directIncludes': '# Inc',
          'directCssIncludes': '# Css Inc',
          'expand__experiences': '# Exp',
          'expand__components': '# Comp',
    }

    handlers = {
          'directIncludes': lambda x: str(len(x)),
          'directCssIncludes': lambda x: str(len(x)),
          'expand__experiences': lambda x: str(len(x)),
          'expand__components': lambda x: str(len(x))
    }

    resp = self.util.cli_request('GET',
        self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions?expand=experiences&page={page}&pagesize={pagesize}&{filter}',
        {'component_set_id': component_set_id, **data}))

    if not self.json:
      self.util.print_table(resp['payload']['data'], self.format, titles=titles, handlers=handlers)
      self.util.print_record_count(resp)
    else:
      click.echo(json.dumps(resp))

  def list_components(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')

    version = self.util.get_component_set_version(component_set_id, self.version)
    version_id = version.get('id')

    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions/{version_id}?expand=components',\
      {'component_set_id': component_set_id, 'version_id': version_id}))

    comps = resp.get('payload', {}).get('data', {}).get('expand')
    click.echo(json.dumps(comps))

  def create(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')

    if self.from_version:
      version_data = self.util.get_component_set_version(component_set_id, self.from_version)

      if self.version:
        major, minor, patch = self.util.parse_version_number(self.version)

        version_data['major'] = major
        version_data['minor'] = minor
        version_data['patch'] = patch
      else:
        if self.version_bump == 'patch':
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions?major={major}&minor={minor}&sort=patch+desc',\
              { 'component_set_id': version_data['componentSetId'], 'major': version_data['major'], 'minor': version_data['minor'] }))['payload']['data']
        elif self.version_bump == 'minor':
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions?major={major}&sort=minor+desc',\
              { 'component_set_id': version_data['componentSetId'], 'major': version_data['major'] }))['payload']['data']
        else:
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions?sort=major+desc',\
              { 'component_set_id': version_data['componentSetId'] }))['payload']['data']

        if versions:
          latest_version = versions[0]

        version_data[self.version_bump] = latest_version[self.version_bump] + 1
        if self.version_bump in ['major', 'minor']:
          version_data['patch'] = 0
        if self.version_bump == 'major':
          version_data['minor'] = 0

    else:
      version_data = {}
      if not self.label:
        self.label = click.prompt(click.style("Label", fg="white"), type=click.Choice(['prod', 'dev', 'old']))

      if not self.version:
        self.version = click.prompt(click.style("Version Number", fg="white"), type=str)

    if self.version:
      major, minor, patch = self.util.parse_version_number(self.version)

      version_data['major'] = major
      version_data['minor'] = minor
      version_data['patch'] = patch

    if not self.json:
      click.echo('Image Size: {}'.format(self.util.get_size(self.component_set_file_path, image=False)))
      click.echo("Uploading Component Set Version to Lumavate")

    version_data['ephemeralKey'] = self.util.upload_ephemeral(self.component_set_file_path, 'application/zip')

    if self.label:
      version_data['label'] = self.label

    if self.css_includes:
      if 'directCssIncludes' not in version_data.keys():
        version_data['directCssIncludes'] = [x for x in self.css_includes]
      else:
        version_data['directCssIncludes'].extend([x for x in self.css_includes])

    if self.direct_includes:
      if 'directIncludes' not in version_data.keys():
        version_data['directIncludes'] = [x for x in self.direct_includes]
      else:
        version_data['directIncludes'].extend([x for x in self.direct_includes])

    resp = self.util.cli_request('POST',
        self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions',
        {'component_set_id': component_set_id}), data=json.dumps(version_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    handlers = {
      'directIncludes': lambda x: str(len(x)),
      'directCssIncludes': lambda x: str(len(x))
    }

    self.util.print_table([resp['payload']['data']], self.format, handlers=handlers)

  def update(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')

    version = self.util.get_component_set_version(component_set_id, self.version)
    version_id = version.get('id')

    post_data = {
      'componentSetId': component_set_id,
      'id': version_id,
      'label': self.label
    }

    resp = self.util.cli_request('PUT',
      self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions/{version_id}',
      {'component_set_id': component_set_id, 'version_id': version_id}), data=json.dumps(post_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    handlers = {
      'directIncludes': lambda x: str(len(x)),
      'directCssIncludes': lambda x: str(len(x))
    }

    self.util.print_table([resp['payload']['data']], self.format, handlers=handlers)

  def delete(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')

    if not self.version:
      self.version = click.prompt(click.style("Version: ", fg="white"), type=str)

    if "*" in self.version:
      versions = self.util.get_versions_from_mask('component-set', component_set_id, self.version)
      if self.json:
        resp_list = {}
        resp_list['responses'] = []
        for v in versions:
          resp = self.util.cli_request('DELETE',
                self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions/{version_id}', {'component_set_id': component_set_id, 'version_id': v['id']}))
          resp_list['responses'].append(resp)

        click.echo(json.dumps(resp_list))
        return

      with click.progressbar(versions, label='Deleting Versions') as bar:
        for v in bar:
          resp = self.util.cli_request('DELETE',
              self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions/{version_id}', {'component_set_id': component_set_id, 'version_id': v['id']}))

      handlers = {
          'directIncludes': lambda x: str(len(x)),
          'directCssIncludes': lambda x: str(len(x))
      }

      self.util.print_table(versions, self.format, handlers=handlers)

    else:
      version = self.util.get_component_set_version(component_set_id, self.version)
      version_id = version.get('id')

      resp = self.util.cli_request('DELETE',
          self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/versions/{version_id}',
          {'component_set_id': component_set_id, 'version_id': version_id}))

      if self.json:
        click.echo(json.dumps(resp))
        return

      handlers = {
        'directIncludes': lambda x: str(len(x)),
        'directCssIncludes': lambda x: str(len(x))
      }

      self.util.print_table([resp['payload']['data']], self.format, handlers=handlers)



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import click

@click.group(help='A widget is comprised of widget-versions. You can share widgets with other orgs giving them access to use the associated versions.')
def widget():
    pass


import commands
import behavior
import click
import pyke

@click.command('add', help='Create a new version of a container')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--port', type=int, help='The port to expose on your container')
@click.option('--editor-port', type=int, help="The port to connect to the container's editor")
@click.option('--is-editable', is_flag=True, help="Set if this is an editable container. Must come from a version with an editor port")
@click.option('--docker-image', '-image', type=str, help='The name (including the tag) of a docker image from the docker daemon. You must have docker running locally to use this option.')
@click.option('--container-file-path', '-path', type=click.Path(), help='The image must be a gzipped tar file. Ex: {file_name}.tar.gz')
@click.option('--from-version', '-fv', type=str, help=pyke.helptxt.FROM_VERSION)
@click.option('--version', '-v', type=str, default=None, help="The version number of the new version")
@click.option('--patch', 'version_bump', flag_value='patch', default=True)
@click.option('--minor', 'version_bump', flag_value='minor')
@click.option('--major', 'version_bump', flag_value='major')
@click.option('--env', '-e', default=None, multiple=True, help="The environment variables to add to the docker image when it is run -- Syntax: envVar=value")
@click.option('--label', '-l', type=click.Choice(['prod', 'dev', 'old']))
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {label} {isEditable} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def create(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).create()

@click.command('ls', help='List versions for a container')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {label} {isEditable} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def list(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).list()

@click.command('start', help='Start a container version')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def start(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).start_stop()

@click.command('stop', help='Stop a container version')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--format', '-f', default='{id} {actualState} {versionNumber} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def stop(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).start_stop()

@click.command('exec', help='Pass in a command to be run directly on the docker container.')
@click.argument('command')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def run(command, **kwargs):
  behavior.ContainerVersionBehavior(**kwargs).run(command)

@click.command(help='Update a container version')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--label', '-l', prompt=True)
@click.option('--format', '-f', default='{id} {versionNumber} {label} {isEditable} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def update(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).update()

@click.command('rm', help='Delete a container version')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True)
@click.option('--version', '-v', prompt=True, help='The version ID, the version number or a version mask')
@click.option('--format', '-f', default='{id} {versionNumber} {label} {isEditable} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def delete(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).delete()

@click.command('logs', help='Tail logs for an editable container')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The ID or the url-ref of the container associated with the editable version')
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--tail-number', '-n', type=int, help='The number of lines to show')
@click.option('--editor', is_flag=True, help='Return raw json from the platform')
def tail(editor, **kwargs):
  behavior.ContainerVersionBehavior(**kwargs).tail_logs(editor)

@click.command('editorInfo', help='Get version info from the editor')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The ID or the url-ref of the container associated with the editable version')
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--tail-number', '-n', type=int, help='The number of lines to show')
def editor_info(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).editor_info()

@click.command(help='This force updates your docker service. This can recover a process in a bad state')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The ID or the url-ref of the container associated with the editable version')
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
def restart(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).force_update()

@click.command('download', help='Download a zip file that contains this containers application source code')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The ID or the url-ref of the container associated with the editable version')
@click.option('--version', '-v', prompt=True, help='The version ID or the version-number')
@click.option('--path', '-path', type=click.Path(), help='Path to save zip file')
def download_app_zip(**kwargs):
  behavior.ContainerVersionBehavior(**kwargs).download_app_zip()


commands.microservice_version.add_command(create)
commands.widget_version.add_command(create)

commands.microservice_version.add_command(list)
commands.widget_version.add_command(list)

commands.microservice_version.add_command(start)
commands.widget_version.add_command(start)

commands.microservice_version.add_command(stop)
commands.widget_version.add_command(stop)

commands.microservice_version.add_command(delete)
commands.widget_version.add_command(delete)

commands.microservice_version.add_command(update)
commands.widget_version.add_command(update)

commands.microservice_version.add_command(run)
commands.widget_version.add_command(run)

commands.microservice_version.add_command(download_app_zip)
commands.widget_version.add_command(download_app_zip)

commands.microservice_version.add_command(restart)
commands.widget_version.add_command(restart)

commands.microservice_version.add_command(tail)
commands.widget_version.add_command(tail)

commands.microservice_version.add_command(editor_info)
commands.widget_version.add_command(editor_info)



from pyparsing import Literal, Word, nums, Combine, Optional,\
  delimitedList, oneOf, alphas, Suppress
import dateutil.parser
import click
import pyke
import re

ESC = Literal('\x1b')
integer = Word(nums)
escapeSeq = Combine(ESC + '[' + Optional(delimitedList(integer,';')) +
                oneOf(list(alphas)))

uncolor = lambda s : Suppress(escapeSeq).transformString(s)

class Menu:
  def __init__(self, json=False):
    self.json = json

  def camel_to_underscore(self, name):
    camel_pat = re.compile(r'([A-Z0-9])')
    click.echo(camel_pat.sub(lambda x: ' ' + x.group(1).lower(), name))
    return camel_pat.sub(lambda x: ' ' + x.group(1).lower(), name)

  def underscore_to_camelcase(self, value):
    def camelcase():
      yield str.lower
      while True:
        yield str.capitalize

    c = camelcase()
    return "".join(c.next()(x) if x else '_' for x in value.split("_"))

  def sanatize_data(self, data, skip=None):
    if not skip:
      skip = ['orgName', 'name']

    data = { k:v.replace(" ", "") if isinstance(v, str) and k not in skip else v.strip() if k == 'orgName' or k =='name' else v for k,v in data.items() }
    if 'Option' in data.keys():
      del data['Option']

    return data

  def _process_data(self, data, sort=None):
    if sort:
      data = sorted(data, key=lambda k: (sort not in k, k.get(sort, None)))

    # Add option Column
    ct = 1
    for rec in data:
      rec['Option'] = ct
      ct = ct + 1

    return data

  def get_user_select(self, data, format, sort=None, titles=None, skip=None):
    if '{Option}' not in format:
      format = '{Option} ' + format
    data = self._process_data(data, sort)
    data = self.print_menu(data, format, titles)
    return self._get_selection(data, skip=skip)

  def print_menu(self, data, format, titles={}, handlers={}):
    if not titles:
      titles = {}

    headers = {}
    stats = {}

    for c in data:
      child_records = {}
      for k in c:
        if isinstance(c[k], dict):
          for sk in c[k]:
            child_records[k + '__' + sk] = c[k][sk]

      c.update(child_records)

    if len(data) > 0:
      for k in data[0]:
        headers[k] = titles.get(k, k)
    else:
      click.echo('No Records Found')
      return

    for k in headers:
      stats[k] = len(uncolor(headers[k]))

    for c in data:
      for k in c:
        if k in ['createdAt', 'lastModifiedAt']:
          cd = dateutil.parser.parse(c[k])
          c[k] = cd.strftime('%x %X')
        if stats.get(k, 0) < len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k])):
          stats[k] = len(str(uncolor(handlers[k](c[k])) if k in handlers else c[k]))

    for k in headers:
      headers[k] = titles.get(k, k).ljust(stats[k])

    for c in data:
      for k in c:
        c[k] = str(handlers[k](c[k]) if k in handlers else c[k])
        c[k] = c[k] + ' ' * (stats[k] - len(uncolor(c[k])))

    try:
      col_headers = format.format(**headers)
    except KeyError as ke:
      raise click.ClickException(click.style(f'You passed in a header that did not exsist in the response: {ke}', fg='red'))

    click.echo(click.style(col_headers, fg='white'))

    for c in data:
      click.echo(format.format(**c))

    click.echo(" ")

    return data

  def validate_num(self, val, min_val=None):
    try:
      int_val = int(val)
      if min_val:
        if int_val < min_val:
          raise click.ClickException(click.style(f"Expected an interger greater than {min_val} as input", fg='red'))

      return int_val

    except:
      if min_val:
        raise click.ClickException(click.style(f"Expected an interger greater than {min_val} as input", fg='red'))
      else:
        raise click.ClickException(click.style("Expected an interger as input", fg='red'))

  def _get_selection(self, data, skip=None):
    user_num = click.prompt(click.style("Enter the Option #", fg="white"), type=int)
    click.echo(' ')

    selection = None
    for rec in data:
      rec_num = rec.get('Option').strip()
      user_num = str(user_num).strip()

      if rec_num == user_num:
        selection = rec
        break

    if not selection:
      raise click.ClickException(click.style(f"{user_num} was not in the list of options", fg='red'))

    return self.sanatize_data(selection, skip=skip)



import commands
import behavior
import click
import pyke


@click.command('ls')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
def list(**kwargs):
    behavior.ContainerBehavior(**kwargs).list()

@click.command('rm')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The id or the urlRef of the container.')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
def delete(**kwargs):
    behavior.ContainerBehavior(**kwargs).delete()

@click.command('add')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--name', prompt=True, help='The name of the container to create.')
@click.option('--url-ref', prompt=True, help="A unique value for how the container will be addressed.")
@click.option('--icon-file', '-path', type=click.Path(), help='The path to an SVG icon for the container.')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
def create(**kwargs):
    behavior.ContainerBehavior(**kwargs).create()

@click.command('access', help="The access command shares and unshares objects with child orgs.")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The id or the urlRef of the container.')
@click.option('--add', multiple=True, help="Share with an org. (ID || Org Name)")
@click.option('--rm', multiple=True, help="Un-Share with an org. (ID || Org Name)")
@click.option('--absolute', multiple=True, help='Flush and fill shared with list. (ID || Org Name)')
@click.option('--format', '-f', default='{failed} {sharedWith} {unsharedFrom} {resultingGrantees}', help=pyke.helptxt.FORMAT)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
@click.option('--current', is_flag=True, help="Returns a list of orgs currently shared with this object.")
def share(**kwargs):
    behavior.ContainerBehavior(**kwargs).share()

@click.command()
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--container', '-c', prompt=True, help='The id or the urlRef of the container.')
@click.option('--name', help='A new name for the container.')
@click.option('--icon-file', '-path', type=click.Path(), help='The path to a new SVG icon for the container.')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json/--table', default=False, help='Return raw json from the platform.')
def update(**kwargs):
    behavior.ContainerBehavior(**kwargs).update()


commands.microservice.add_command(list)
commands.widget.add_command(list)

commands.microservice.add_command(share)
commands.widget.add_command(share)

commands.widget.add_command(create)
commands.microservice.add_command(create)

commands.widget.add_command(delete)
commands.microservice.add_command(delete)

commands.widget.add_command(update)
commands.microservice.add_command(update)



"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

import pyqrcode as qrc
from cli import pyke
import requests
import urllib3
import click
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

class ProjectBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')

  def read_config(self):
    pass

  def write_config(self):
    pass

  def init(self):
    ts = pyke.Tarsum()
    tarfile = ts.make_tarfile('mon_tar.tar', '/home/dev/Code/monitor')
    ts.gzip_file(tarfile)
    click.echo(ts.get_checksum(f'{os.getcwd()}/mon_tar.tar'))

  def display_qr(self):
    url = qrc.create('https://google.com')
    url.svg('qr-activation.svg', scale=8)
    click.echo(url.terminal(quiet_zone=2))



"""
    These sub-commands and their corresponding logic are both implamented in this file.
    TO DO: Break out the logic for these commands into a seperate behavior file.
"""

from pathlib import Path
import requests
import urllib3
import click
import json
import pyke
import sys
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
  urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
else:
  verify_tls = None

@click.group(help="Profiles add company context to commands. Almost all calls to Lumavate require context. \
Profiles use Envs to get tokens and then give those tokens context using the Org ID.")
def profile():
    pass

@click.command('ls', help="List the profiles you have configured.")
@click.option('--format', '-f', default='{profileName} {env} {orgName} {orgId}', help='The --format option takes the column name of the returned table wrapped in \
{} and returns only that column. It is not compatible with --json flag.')
@click.option('--json', 'json_flag', is_flag=True, help='Return raw json.')
def list(format, json_flag):
  profile_data = pyke.auth.list_profiles()
  if json_flag:
    click.echo(json.dumps(profile_data))
    return

  titles = {"profileName": "Profile Name", "env": "Env", "orgName": "Org Name", "orgId": "Org ID"}
  pyke.Menu().print_menu(profile_data, format, titles=titles)

@click.command()
@click.option('--profile-name', help='Profile Name')
@click.option('--env', default=None, help='The name of the environment you want to use with this profile')
@click.option('--format', '-f', default='{profileName} {env} {orgName} {orgId}', help='The available column names for this command are {env} {orgName} {orgId}')
def add(profile_name, env, format):
  if not profile_name:
    profile_name = click.prompt(click.style("Profile Name", fg="white"), type=str)

  click.echo(' ')

  titles = {
      'envName': 'Env Name',
      'app': 'App',
      'audience': 'Audience',
      'token': 'Token'
  }

  headers = '{envName} {app} {audience} {token}'

  if env:
    env_data = pyke.auth.get_env_data(env)
  else:
    envs = pyke.auth.list_envs()
    env_data = pyke.Menu().get_user_select(envs, headers, sort='envName', titles=titles)

  util = pyke.Util(ctx=env_data)

  titles = {
      'name': 'Org Name',
      'instanceType': 'Org Type',
      'isTest': 'Test Org'
    }
  headers = "{name} {instanceType} {isTest}"
  company = pyke.Menu().get_user_select(util.get_companies(), headers, sort='orgName', titles=titles)

  profile_data = {}
  profile_data["orgId"] = company.get('id')
  profile_data["orgName"] = company.get('name')
  profile_data["env"] = util.ctx.get('envName')
  profile_data["instanceType"] = company.get('instanceType')
  profile_data["profileName"] = profile_name

  pyke.auth.update_profile(profile_name, profile_data)
  util.ctx.update(profile_data)

  company_context = get_company_context(util.ctx).get('company')

  if company_context:
    profile_data = pyke.auth.get_profile_data(profile_name)
    profile_data['experienceCloudUri'] = company_context['experienceCloudUri']
    pyke.auth.update_profile(profile_name, profile_data)

  else:
    raise click.ClickException(click.style('Error getting company context', fg='red'))

  titles = {
      'profileName': 'Profile',
      'env': 'Environment',
      'orgName': 'Org Name',
      'orgId': 'Org ID'
  }

  pyke.Menu().print_menu([profile_data], format, titles=titles)

def get_company_context(ctx):
  util = pyke.Util(ctx=ctx)
  token_resp = util.login()
  headers = {
        'Authorization': 'Bearer ' + token_resp.get('access_token'),
        'Content-Type': 'application/json'
      }

  resp = requests.get(util.build_url('{app}/auth/v1/context'), headers=headers, verify=verify_tls)

  if resp.status_code == 200:
    return resp.json()['payload']['data']
  else:
    raise click.ClickException(click.style('Error getting company context', fg='red'))

@click.command('refresh-token', help="Refreshes the access token stored in the config - Adds 'experienceCloudUri' and 'instanceType' to the profile if they are not present." )
@click.option('--profile', '-p', help='Profile to refresh.', callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
def refresh_token(profile):
  profile.login()
  profile_name = profile.ctx.get('profileName')
  click.echo(f"Refreshed '{profile_name}' access token.")

  # Migration path for old profile data without experienceCloudUri
  if 'experienceCloudUri' not in profile.ctx.keys():
    company_context = get_company_context(profile.ctx).get('company')

    if company_context:
      profile_data = pyke.auth.get_profile_data(profile_name)
      profile_data['experienceCloudUri'] = company_context.get('experienceCloudUri')
      pyke.auth.update_profile(profile_name, profile_data)

  if 'instanceType' not in profile.ctx.keys():
    company_context = get_company_context(profile.ctx).get('company')

    if company_context:
      profile_data = pyke.auth.get_profile_data(profile_name)
      profile_data['instanceType'] = company_context.get('instanceType')
      pyke.auth.update_profile(profile_name, profile_data)

@click.command('rm')
@click.argument('profile')
def delete(profile):
  config_data = pyke.auth.load_config()
  profile_data = config_data["profiles"].get(profile)
  if not profile_data:
    raise click.ClickException(click.style(f"Could not find the profile {profile} to delete", fg='red'))

  del config_data["profiles"][profile]
  pyke.auth.save_config(config_data)
  click.echo(click.style(f"Deleted profile: {profile}", fg="white"))

profile.add_command(list)
profile.add_command(delete)
profile.add_command(add)
profile.add_command(refresh_token)



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import click

@click.group(help='A microservice is comprised of microservice-versions. You can share microservices with other orgs giving them access to use the associated versions.')
def microservice():
    pass


from datetime import datetime
from pathlib import Path
import subprocess
import platform
import click
import pyke
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
else:
  verify_tls = None

class ContainerBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')

    click_ctx           = click.get_current_context()
    parent_command      = click_ctx.parent.command.name
    self.container_type = parent_command.split("-")[0]

    self.format         = kwargs.get('format')
    self.json           = kwargs.get('json')
    self.container      = kwargs.get('container')
    self.name           = kwargs.get('name')
    self.url_ref        = kwargs.get('url_ref')
    self.icon_file      = kwargs.get('icon_file')
    self.add            = kwargs.get('add')
    self.rm             = kwargs.get('rm')
    self.absolute       = kwargs.get('absolute')
    self.current        = kwargs.get('current')
    self.filter         = kwargs.get('filter')
    self.page           = kwargs.get('page')
    self.pagesize       = kwargs.get('pagesize')

  def list(self):
    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize, 'container_type': self.container_type}

    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/containers?type={container_type}&page={page}&pagesize={pagesize}&{filter}', {**data} ))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table(resp['payload']['data'], self.format)
    self.util.print_record_count(resp)

  def create(self):
    # Replace with os.walk()
    if not self.icon_file:
      self.icon_file = self.util.resolve_default_icon_path()

    post_data = {
      'name': self.name,
      'urlRef': self.url_ref,
      'type': self.container_type,
      'ephemeralKey': self.util.upload_ephemeral(self.icon_file, 'image/svg+xml'),
      'publisherName': self.util.ctx['orgName'],
    }

    resp = self.util.cli_request('POST',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers'), data=json.dumps(post_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    pyke.Menu().print_menu([resp['payload']['data']], self.format)

  def delete(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    versions = self.util.get_all_container_versions(container_id)
    if versions:
      raise click.ClickException(click.style("You cannont delete a container with versions", fg='red'))

    resp = self.util.cli_request('DELETE',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}', {'container_id': container_id}))

    if self.json:
      click.echo(json.dumps(resp))
      return

    pyke.Menu().print_menu([resp['payload']['data']], self.format)

  def update(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')
    post_data = {}

    if not self.name and not self.icon_file:
      click.echo(click.style("You must provide either a new name or new icon to update the container", fg='yellow'))
      return

    if self.name:
      post_data['name'] = self.name

    if self.icon_file:
      post_data['ephemeralKey'] = self.util.upload_ephemeral(self.icon_file, 'image/svg+xml')

    resp = self.util.cli_request('PUT',
      self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}', {'container_id': container_id}),
      data=json.dumps(post_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table([resp['payload']['data']], self.format)

  def share(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    if self.current:
      if self.json:
        resp = self.util.cli_request('GET',
          self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/access', {'container_id': container_id}))

        click.echo(resp)
        return
      else:
        resp = self.util.cli_request('GET',
          self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

        recs = self.util.get_current_access_records('container', container_id)
        org_names = [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]
        click.echo(org_names)
        return

    resp = self.util.cli_request('GET',
        self.util.build_url('{app}/auth/v1/company/child-companies'))['payload']['data']

    handlers = {
        'sharedWith': lambda x: x.get('granteeId') if x is not None else '',
        'unsharedWith': lambda x: x.get('granteeId') if x is not None else ''
      }

    if self.absolute:
      absolute_ids = [self.util.get_company_id(x, resp) for x in self.absolute]
      data = {
          "id": container_id,
          "granteeCompanyIds": [x for x in absolute_ids]
        }

      access_resp = self.util.cli_request('PUT', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/access',\
          {'container_id': container_id}), data=json.dumps(data))

      if self.json:
        click.echo(json.dumps(access_resp))
        return

      recs = self.util.get_current_access_records('container', container_id)
      access_resp['payload']['data']['resultingGrantees'] =\
          [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]

      self.util.print_table([access_resp['payload']['data']], self.format)
      return

    recs = self.util.get_current_access_records('container', container_id)
    sharees = [x.get('granteeCompanyId') for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]
    if self.add:
      add_ids = [self.util.get_company_id(x, resp) for x in self.add]
      sharees.extend(add_ids)

    if self.rm:
      rm_ids = [self.util.get_company_id(x, resp) for x in self.rm]
      sharees = [x for x in sharees if x not in rm_ids]

    data = {
      "id": container_id,
      "granteeCompanyIds": sharees
    }

    access_resp = self.util.cli_request('PUT',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/access', {'container_id': container_id}), data=json.dumps(data))

    if self.json:
      click.echo(json.dumps(access_resp))
      return

    recs = self.util.get_current_access_records('container', container_id)

    access_resp['payload']['data']['resultingGrantees'] =\
        [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]

    self.util.print_table([access_resp['payload']['data']], self.format)



class helptxt:
  PROFILE   = 'The name of the profile to use as the context for the command.'
  FORMAT    = 'The --format option takes the column name of the returned table wrapped in {} and returns only that column. It is not compatible with --json flag.'
  JSON_FLAG = 'Return raw json from the platform.'
  FROM_VERSION = "Take all values from an existing version and create a new version. \
Increment the major, minor or patch version by 1 with --major, --minor, --patch flags. Override any old values by passing them in as options."



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import behavior
import click
import pyke

@click.group(help='Experience Collection related commands.')
def experience_collection():
  pass

@click.command('ls', help="List the experience collections related to the selected profile.")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {createdAt}', help='The --format option takes the column name of the returned table wrapped in \
  {} and returns only that column. It is not compatible with --json flag.')
@pyke.list_options
def list(**kwargs):
  behavior.ExperienceCollectionBehavior(**kwargs).list()

experience_collection.add_command(list)



"""
    The primary use of the ContainerUtil class is to connect to and communicate with
    the running docker daemon.
"""

from pathlib import Path
import docker
import shutil
import click
import gzip
import time
import sys
import os

class ContainerUtil:
  def __init__(self):
    try:
      self.client = docker.DockerClient(base_url='unix://var/run/docker.sock')
    except Exception as e:
      click.echo(click.style("When this error happens it's usualy because Docker is not running, OR there are leftover \
        references to the pip docker-py package on the machine.", fg='yellow'))
      click.echo(click.style("If it's the latter, make sure to remove all references of docker-py. \
        These can often persist in multiple versions of python on your machine even after a pip uninstall.", fg='yellow'))
      click.echo(" ")
      raise click.ClickException(click.style("Error communicating with the docker daemon. {}".format(e), fg='red'))

  def save_image(self, image_name):
    try:
      image = self.client.images.get(image_name)
    except Exception as e:
      raise Exception(click.style('Error getting image: {}'.format(e), fg='red'))

    file_path = '{}/{}'.format(str(Path.home()), image_name.replace('/', '_').replace('.', '_').replace(':', '_'))

    with open(file_path, 'wb') as image_tar:
      for chunk in image.save(named=True):
        image_tar.write(chunk)

    zipped_file_path = self.zip_image(file_path)

    # Cleanup
    os.remove(file_path)

    return zipped_file_path

  def zip_image(self, file_path):
    with open(file_path, 'rb') as f_in:
      with gzip.open('{}.tar.gz'.format(file_path), 'w+b') as f_out:
        shutil.copyfileobj(f_in, f_out)

    return '{}.tar.gz'.format(file_path)




"""
    The auth class is resposible for storing, reading and updating tokens and config settings.
    The CLI should cache tokens until a new one is required. If that's the case the auth class
    retrieves the config variables so a new token can be requested from the platform. Upon success
    the new token is cached and used for subsequent commandsself.
"""

from functools import wraps
from pathlib import Path
import dateutil.parser
from .menu import Menu
import functools
import click
import copy
import json
import os


class auth:

  @staticmethod
  def list_profiles():
    profile_data = auth.load_config().get('profiles')
    for profile_name in profile_data.keys():
      profile_data[profile_name]['profileName'] = profile_name

      # Migrate to new authUrl
      if 'experienceCloudUri' in profile_data[profile_name].keys() and 'authUri' not in profile_data[profile_name].keys():
        profile_data[profile_name]['authUri'] = profile_data[profile_name].get('experienceCloudUri').replace('experience', 'auth')
        auth.update_profile(profile_name, profile_data)

    profiles = [profile_data.get(k) for k in profile_data.keys()]
    return profiles

  @staticmethod
  def list_envs():
    env_data = auth.load_config().get('envs')
    for env_name in env_data.keys():
      env_data[env_name]['envName'] = env_name

    envs = [env_data.get(k) for k in env_data.keys()]
    return envs

  @staticmethod
  def update_profile(name, profile_data):
    config = auth.load_config()
    profiles = config['profiles']

    if profiles.get(name):
      profiles[name].update(profile_data)
    else:
      profiles[name] = profile_data

    auth.save_config(config)
    return config

  @staticmethod
  def get_profile(config_data, name):
    return config_data['profiles'].get(name)

  @staticmethod
  def get_profile_data(key):
    config_data = auth.load_config()
    profile_data = config_data['profiles'].get(key)
    return Menu().sanatize_data(profile_data)

  @staticmethod
  def get_env_data(key):
    config_data = auth.load_config()
    env_data = config_data['envs'].get(key)
    if not env_data:
      raise click.ClickException(click.style(f"Could not find an environment named {key}", fg="red"))
    return Menu().sanatize_data(env_data)

  @staticmethod
  def get_env(config_data, name):
    return config_data['envs'].get(name)

  @staticmethod
  def get_profile_names(ctx, args, incomplete):
    config = auth.load_config()
    names = config['profiles'].keys()
    return [k for k in names if incomplete in k]

  @staticmethod
  def sanatize_data(data):
    skip = ['orgName', 'env', 'envName']
    return { k:v.replace(" ", "") if isinstance(v, str) and k not in skip else v for k,v in data.items() }

  @staticmethod
  def set_profile_token(profile, token):
    profile_data = auth.get_profile_data(profile)
    profile_data['accessToken'] = token
    auth.update_profile(profile, profile_data)

  @staticmethod
  def save_config(data):
    for k in data['envs'].keys():
      data['envs'][k] = auth.sanatize_data(data['envs'][k])

    with open(auth.get_config_file(), 'w+') as outfile:
      json.dump(data, outfile, indent=4, sort_keys=True)

  @staticmethod
  def load_config():
    config_path = auth.get_config_file()
    try:
      with open(config_path) as json_file:
        data = json.load(json_file)
      return data
    except Exception as e:
      raise click.ClickException(click.style('Bad config file found. Please manually fix your congif. Path: {}'.format(config_path), fg='red'))

  @staticmethod
  def get_config_file():
    config_path = f"{Path.home()}/.luma/luma_cli_config.json"
    if not os.path.exists(config_path):
      with open(config_path, 'w') as config:
        json.dump({ "envs": {}, "profiles": {} }, config)

    return config_path



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import behavior
import click
import pyke

@click.group('component-set', help='A component-set is comprised of component-set-versions. You can share component-sets with other orgs giving them access to use the associated versions.')
def component_set():
    pass

@click.command('ls')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def list(**kwargs):
    behavior.ComponentSetBehavior(**kwargs).list()

@click.command('rm')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to delete')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def delete(**kwargs):
    behavior.ComponentSetBehavior(**kwargs).delete()

@click.command('add')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--name', prompt=True, help='The name of the component set')
@click.option('--url-ref', prompt=True, help="A unique value for how the component-set will be addressed in a URL.")
@click.option('--icon-file', '-path', type=click.Path(file_okay=True, dir_okay=False, exists=True), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def create(**kwargs):
    behavior.ComponentSetBehavior(**kwargs).create()

@click.command()
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to update')
@click.option('--name', default='', help='The name of the component set')
@click.option('--icon-file', '-path', type=click.Path(file_okay=True, dir_okay=False, exists=True), help='The path to an SVG icon for the component set')
@click.option('--format', '-f', default='{id} {name} {urlRef} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def update(**kwargs):
    behavior.ComponentSetBehavior(**kwargs).update()

@click.command('access', help="The access command shares and unshares objects with child orgs.")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set being shared/unshared')
@click.option('--add', multiple=True, help="Share with an org. (ID || Org Name)")
@click.option('--rm', multiple=True, help="Un-Share with an org. (ID || Org Name)")
@click.option('--absolute', multiple=True, help='Flush and fill shared with list. (ID || Org Name)')
@click.option('--format', '-f', default='{failed} {sharedWith} {unsharedFrom} {resultingGrantees}', help=pyke.helptxt.FORMAT)
@click.option('--current', is_flag=True, help="Returns a list of orgs currently shared with this object.")
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def share(**kwargs):
    behavior.ComponentSetBehavior(**kwargs).share()

component_set.add_command(list)
component_set.add_command(create)
component_set.add_command(delete)
component_set.add_command(update)
component_set.add_command(share)



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import click

@click.group('microservice-version', help="A microservice-version is the docker image of a non visual widget, \
along with the details it needs to run such as, port number and Env variables.")
def microservice_version():
    pass


"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import behavior
import click
import pyke

@click.group('component-set-version', help="A component-set-version is where the platform looks for components when used in a widget. \
The version also contains other details like CSS includes.")
def component_set_version():
    pass

@click.command('ls')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set')
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {expand__experiences} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--filter', default='')
@click.option('--page', default=1)
@click.option('--pagesize', default=100)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def list(**kwargs):
    behavior.ComponentSetVersionBehavior(**kwargs).list()

@click.command('components', help="Given a --component-set and a --version, this will return the raw JSON contained in a component-set-version")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The id or the urlRef of the component-set associated with the version')
@click.option('--version', '-v', prompt=True)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def components(**kwargs):
    behavior.ComponentSetVersionBehavior(**kwargs).list_components()

@click.command('add', help="While adding a new component-set-version you must always upload a zipped file, even if you use the from-version option")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set to associate with the new version')
@click.option('--component-set-file-path', '-path', prompt=True, required=True, type=click.Path(file_okay=True, dir_okay=False, exists=True), help='Must be the path to a zipped file.')
@click.option('--from-version', '-fv', type=str, help=pyke.helptxt.FROM_VERSION)
@click.option('--version', '-v', type=str, default=None, help="The version number of the new version")
@click.option('--patch', 'version_bump', flag_value='patch', default=True)
@click.option('--minor', 'version_bump', flag_value='minor')
@click.option('--major', 'version_bump', flag_value='major')
@click.option('--css-includes', type=str, default=None, multiple=True)
@click.option('--direct-includes', type=str, default=None, multiple=True)
@click.option('--label', '-l', type=click.Choice(['prod', 'dev', 'old']))
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def create(**kwargs):
    behavior.ComponentSetVersionBehavior(**kwargs).create()

@click.command()
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set associated with the version')
@click.option('--version', '-v', prompt=True, help='The ID or the version number of the version to update')
@click.option('--label', '-l', prompt=True, type=click.Choice(['prod', 'dev', 'old']))
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def update(**kwargs):
    behavior.ComponentSetVersionBehavior(**kwargs).update()

@click.command('rm')
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
@click.option('--component-set', '-cs', prompt=True, help='The ID or the urlRef of the component-set associated with the version')
@click.option('--version', '-v', prompt=True, help='The version ID, the version number or a version mask')
@click.option('--format', '-f', default='{id} {versionNumber} {directIncludes} {directCssIncludes} {label} {createdAt}', help=pyke.helptxt.FORMAT)
@click.option('--json', is_flag=True, help='Return raw json from the platform')
def delete(**kwargs):
    behavior.ComponentSetVersionBehavior(**kwargs).delete()

component_set_version.add_command(list)
component_set_version.add_command(create)
component_set_version.add_command(delete)
component_set_version.add_command(update)
component_set_version.add_command(components)



"""
    These are the sub-commands for a top level object added to the CLI group in 'cli.py'.
    The commands and options are implemented here and the logic behind them resides in the corresponding behavior file.
"""

import behavior
import click
import pyke

@click.group(help='Commands related to developing, managing and publishing experiences.')
def project():
  pass

@click.command('init', help="Initializes the current directory as a lumavate project")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
def init(**kwargs):
  #behavior.ProjectBehavior(**kwargs).init()
  pass

@click.command('get-qr', help="Display the QR activation code in the terminal")
@click.option('--profile', '-p', help=pyke.helptxt.PROFILE, callback=pyke.select_profile, autocompletion=pyke.auth.get_profile_names)
def get_qr(**kwargs):
  #behavior.ProjectBehavior(**kwargs).display_qr()
  pass


project.add_command(init)
project.add_command(get_qr)



"""
    This behavior file contains the logic for a subset of commands. Logic specific to
    commands should be implamented in corresponding behavior files.
"""

from pathlib import Path
import click
import time
import pyke
import json
import sys
import os

class ComponentSetBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')
    self.json           = kwargs.get('json')
    self.component_set  = kwargs.get('component_set')
    self.format         = kwargs.get('format')
    self.filter         = kwargs.get('filter')
    self.page           = kwargs.get('page')
    self.pagesize       = kwargs.get('pagesize')
    self.add            = kwargs.get('add')
    self.rm             = kwargs.get('rm')
    self.absolute       = kwargs.get('absolute')
    self.current        = kwargs.get('current')
    self.name           = kwargs.get('name')
    self.icon_file      = kwargs.get('icon_file')
    self.url_ref        = kwargs.get('url_ref')

  def list(self):
    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/component-sets?page={page}&pagesize={pagesize}&{filter}', {**data} ))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table(resp['payload']['data'], self.format)
    self.util.print_record_count(resp)

  def delete(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')
    versions = self.util.get_all_component_set_versions(component_set_id)
    if versions:
      raise click.ClickException(click.style("You cannont delete a component set with versions", fg='red'))

    resp = self.util.cli_request('DELETE',
        self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}))

    if self.json:
      click.echo(json.dumps(resp))
      return

    pyke.Menu().print_menu([resp['payload']['data']], self.format)

  def create(self):
    if not self.icon_file:
      self.icon_file = self.util.resolve_default_icon_path()

    post_data = {
      'name': self.name,
      'urlRef': self.url_ref,
      'ephemeralKey': self.util.upload_ephemeral(self.icon_file, 'image/svg+xml')
    }

    resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/component-sets'), data=json.dumps(post_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table([resp['payload']['data']], self.format)

  def update(self):
    component_set = self.util.get_component_set(self.component_set)
    component_set_id = component_set.get('id')
    post_data = {}

    if not self.name and not self.icon_file:
      click.echo(click.style("You must provide either a new name or new icon to update the component set", fg='yellow'))
      return

    if self.name:
      post_data['name'] = self.name

    if self.icon_file:
      post_data['ephemeralKey'] = self.util.upload_ephemeral(self.icon_file, 'image/svg+xml')

    resp = self.util.cli_request('PUT',
      self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}', {'component_set_id': component_set_id}),
      data=json.dumps(post_data))

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table([resp['payload']['data']], self.format)

  def share(self):
      component_set = self.util.get_component_set(self.component_set)
      component_set_id = component_set.get('id')

      if self.current:
        if self.json:
          resp = self.util.cli_request('GET',
            self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/access', {'component_set_id': component_set_id}))

          click.echo(resp)
          return
        else:
          resp = self.util.cli_request('GET',
            self.util.build_url('{experienceCloudUri}/auth/v1/company/child-companies'))['payload']['data']

          recs = self.util.get_current_access_records('component-set', component_set_id)
          org_names = [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]
          click.echo(org_names)
          return

      resp = self.util.cli_request('GET',
          self.util.build_url('{experienceCloudUri}/auth/v1/company/child-companies'))['payload']['data']

      handlers = {
          'sharedWith': lambda x: x.get('granteeId') if x is not None else '',
          'unsharedWith': lambda x: x.get('granteeId') if x is not None else ''
        }

      if self.absolute:
        absolute_ids = [self.util.get_company_id(x, resp) for x in self.absolute]
        data = {
            "id": component_set_id,
            "granteeCompanyIds": [x for x in absolute_ids]
          }

        access_resp = self.util.cli_request('PUT', self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/access',\
            {'component_set_id': component_set_id}), data=json.dumps(data))

        if self.json:
          click.echo(json.dumps(access_resp))
          return

        recs = self.util.get_current_access_records('component-set', component_set_id)
        access_resp['payload']['data']['resultingGrantees'] =\
            [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]

        self.util.print_table([access_resp['payload']['data']], self.format)
        return

      recs = self.util.get_current_access_records('component-set', component_set_id)
      sharees = [x.get('granteeCompanyId') for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]
      if self.add:
        add_ids = [self.util.get_company_id(x, resp) for x in self.add]
        sharees.extend(add_ids)

      if self.rm:
        rm_ids = [self.util.get_company_id(x, resp) for x in self.rm]
        sharees = [x for x in sharees if x not in rm_ids]

      data = {
        "id": component_set_id,
        "granteeCompanyIds": sharees
      }

      access_resp = self.util.cli_request('PUT',
          self.util.build_url('{experienceCloudUri}/iot/v1/component-sets/{component_set_id}/access', {'component_set_id': component_set_id}), data=json.dumps(data))

      if self.json:
        click.echo(json.dumps(access_resp))
        return

      recs = self.util.get_current_access_records('component-set', component_set_id)

      access_resp['payload']['data']['resultingGrantees'] =\
          [self.util.get_company_name(x.get('granteeCompanyId'), resp) for x in recs if x.get('granteeCompanyId') != int(self.util.ctx.get('orgId'))]

      self.util.print_table([access_resp['payload']['data']], self.format)



import tarfile
import hashlib
import shutil
import click
import gzip
import os

class Tarsum:
  def __init__(self, target_dir=None):
    self.target_dir = target_dir
    if not self.target_dir:
      self.target_dir = os.getcwd()

  def tar_working_dir(self):
    cwd = self.target_dir
    self.make_tarfile('test.tar.gz', cwd)
    tarfile_path = f'{cwd}/test.tar.gz'

    return tarfile_path

  def check_file(self):
    tarfile_path = self.tar_working_dir()
    stat = os.stat(tarfile_path)
    click.echo("Zipped tarfile size")
    click.echo(stat.st_size)

    checksum = self.get_checksum(tarfile_path)
    click.echo(checksum)

  def get_checksum(self, file_path):
    hash_md5 = hashlib.md5()

    with open(file_path, "rb") as f:
      for chunk in iter(lambda: f.read(4096), b""):
        hash_md5.update(chunk)

    return hash_md5.hexdigest()

  def gzip_file(self, file_path, file_name):
    with open(f'{file_path}/{file_name}', 'rb') as f_in:
      with gzip.open(f'{file_path}/{file_name}.gz', 'wb') as f_out:
        shutil.copyfileobj(f_in, f_out)

    click.echo(dir(f_out))

  def make_tarfile(self, output_filename, source_dir):
    def ignore_project_config(tarinfo):
      if '.lumavate' in tarinfo.path or '.git' in tarinfo.path:
        return None

      return tarinfo

    with tarfile.open(output_filename, "w:") as tar:
      tar.add(source_dir, arcname=os.path.basename(source_dir), filter=ignore_project_config)

    click.echo(dir(tar))
    click.echo(tar.fileobject)
    return tar



from datetime import datetime
from pathlib import Path
import subprocess
import platform
import requests
import click
import pyke
import json
import os

verify_tls = os.environ.get('LUMA_CLI_VERIFY_TLS', None)
if verify_tls == 'false' or verify_tls == 'False':
  verify_tls = False
else:
  verify_tls = None

class ContainerVersionBehavior:
  def __init__(self, **kwargs):
    self.util           = kwargs.get('profile')

    self.click_ctx      = click.get_current_context()
    parent_command      = self.click_ctx.parent.command.name
    self.container_type = parent_command.split("-")[0]

    self.container           = kwargs.get('container')
    self.port                = kwargs.get('port')
    self.editor_port         = kwargs.get('editor_port')
    self.is_editable         = kwargs.get('is_editable')
    self.container_file_path = kwargs.get('container_file_path')
    self.docker_image        = kwargs.get('docker_image')
    self.from_version        = kwargs.get('from_version')
    self.version             = kwargs.get('version')
    self.version_bump        = kwargs.get('version_bump')
    self.env                 = kwargs.get('env', {})
    self.label               = kwargs.get('label')
    self.format              = kwargs.get('format')
    self.json                = kwargs.get('json')
    self.path                = kwargs.get('path')
    self.tail_number         = kwargs.get('tail_number')
    self.filter              = kwargs.get('filter')
    self.page                = kwargs.get('page')
    self.pagesize            = kwargs.get('pagesize')

  def echo(self, msg, color='white'):
    if self.json:
      return
    click.echo(click.style(str(msg), fg=color))

  def raise_exc(self, msg, color='red'):
    raise click.ClickException(click.style(msg, fg=color))

  def list(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')
    data = {'filter': self.filter, 'page': self.page, 'pagesize': self.pagesize}

    handlers = {
      'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
    }

    resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions?page={page}&pagesize={pagesize}&{filter}',
      {'container_id': container_id, **data}))

    if self.json:
      click.echo(json.dumps(resp))
      return

    pyke.Menu().print_menu(resp['payload']['data'], self.format, handlers=handlers)
    self.util.print_record_count(resp)

  def start_stop(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    version_data = self.util.get_container_version(container_id, self.version)
    version_id = version_data.get('id')
    current_state = version_data.get('actualState')

    if current_state not in ['stopped', 'running', 'error']:
      self.raise_exec(f"To start or stop a version it must have a 'stopped', 'running' or 'error' state. The current state of this version is {current_state}. \
        A version goes through several states after being uploaded to Lumavate. It must finish this process before this command can be run. \
        To check its current state, run 'luma container-version ls'.", color='yellow')

    action = self.click_ctx.command.name
    if action == 'stop' and current_state != 'running':
      self.raise_exc("This version is not running", color='yellow')
    if action == 'start' and current_state == 'running':
      self.raise_exc("This version is already running", color='yellow')

    titles = {
        'id': 'id',
        'actualState': 'Current State',
        'versionNumber': 'Version #',
        'createdAt': 'Created At',
        'createdBy': 'Created By'
      }

    data = {
        "action": action,
        "containerId": container_id,
        "id": version_id
      }

    resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/{action}',\
      {'container_id': container_id, 'version_id': version_id, 'action': action}), data=data)

    if self.json:
      click.echo(json.dumps(resp))
      return

    status_id = resp.get('payload', {}).get('data', {}).get('results', {}).get('statusId', {})
    if not status_id:
      return resp

    if action == 'start':
      action_label = 'Starting'
    else:
      action_label = 'Stopping'

    status_resp = self.util.show_progress(status_id, label=f'{action_label} Container')

    error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
    if error_msg is not None:
      self.raise_exc(f'{action_label} failed: {error_msg}')

    click.echo(' ')
    ver_resp = self.util.cli_request('GET',
      self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}', {'container_id': container_id, 'version_id': version_id}))

    handlers = {
      'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
    }

    self.util.print_table([ver_resp['payload']['data']], self.format, handlers=handlers, titles=titles)

  def run(self, command):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')
    version_data = self.util.get_container_version(container_id, self.version)
    version_id = version_data.get('id')

    resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/exec',\
      {'container_id': container_id, 'version_id': version_id}), json={"command": command})

    if self.json:
      click.echo(json.dumps(resp))
      return

    status_id = resp.get('payload', {}).get('data', {}).get('statusId', {})
    if not status_id:
      return resp

    status_resp = self.util.show_progress(status_id, label=f'Executing "{command}""')

    error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
    if error_msg is not None:
      self.raise_exc(f'Command failed: {error_msg}')

    click.echo(' ')
    click.echo('Command Output: ')
    click.echo(' ')
    output = status_resp.get('payload', {}).get('data', {}).get('summary', {}).get('output', {})
    if output:
      for x in output:
        click.echo(x)
    else:
      click.echo(status_resp.get('payload', {}).get('data', {}))

  def update(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')
    version_data = self.util.get_container_version(container_id, self.version)
    version_id = version_data.get('id')

    post_data = {
      'containerId': container_id,
      'id': version_id,
      'label': self.label
      }

    resp = self.util.cli_request('PUT',\
      self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}',
      {'container_id': container_id, 'version_id': version_id}), json=post_data)

    if self.json:
      click.echo(json.dumps(resp))
      return

    self.util.print_table([resp['payload']['data']], self.format)

  def delete(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    if not self.version:
      self.version = click.prompt(click.style("Version: ", fg="white"), type=str)

    if "*" in self.version:
      versions = self.util.get_versions_from_mask('container', container_id, self.version)
      if self.json:
        resp_list = {}
        resp_list['responses'] = []
        for v in versions:
          resp = self.util.cli_request('DELETE',
              self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}', {'container_id': container_id, 'version_id': v['id']}))
          resp_list['responses'].append(resp)

        click.echo(json.dumps(resp_list))
        return

      with click.progressbar(versions, label='Deleting Versions') as bar:
        for v in bar:
          resp = self.util.cli_request('DELETE',
              self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}', {'container_id': container_id, 'version_id': v['id']}))

      self.util.print_table(versions, self.format)
    else:
      version = self.util.get_container_version(container_id, self.version)
      version_id = version.get('id')

      resp = self.util.cli_request('DELETE',
          self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}',
          {'container_id': container_id, 'version_id': version_id}))

      if self.json:
        click.echo(json.dumps(resp))
        return

      self.util.print_table([resp['payload']['data']], self.format)

  def docker_image_path(self, image_name):
    self.echo("Getting docker image:")
    cu = pyke.ContainerUtil()

    return cu.save_image(image_name)

  def resolve_env(self, version):
    if self.env:
      env = {var.split("=", 1)[0]: var.split("=", 1)[1] for var in self.env}
      return env

    return {}

  def create_version_data(self, container_id):
    if not self.from_version:
      version = {}

      # Set label
      if not self.label:
        if self.is_editable:
          label = 'dev'
        else:
          label = input("Label: ")
          if label is None or label == '':
            self.raise_exc("If you don't create from an old version then you must provide a --label")

        version['label'] = label
      else:
        version['label'] = self.label

      # Set version number
      if self.version:
        major, minor, patch = self.util.parse_version_number(self.version)

        version['major'] = major
        version['minor'] = minor
        version['patch'] = patch

      else:
        version_number = input("Version Number: ")

        if not version:
          self.raise_exc("If you don't create from an old version then you must provide a --version-number in the format\
              '<major: int>.<minor: int>.<patch: int>'")

        major, minor, patch = self.util.parse_version_number(version_number)

        version['major'] = major
        version['minor'] = minor
        version['patch'] = patch

      # Set port
      if not self.port:
        port = int(input("Port: "))
        if not port:
          self.raise_exc("If you don't create from an old version then you must provide a --port for your new version")
        version['port'] = port
      else:
        version['port'] = self.port

      return version

    else:
      version = self.util.get_container_version(container_id, self.from_version)
      from_version_id = version.get('id')

      if self.version:
        major, minor, patch = self.util.parse_version_number(self.version)

        version['major'] = major
        version['minor'] = minor
        version['patch'] = patch

        return version

      else:
        if self.version_bump == 'patch':
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions?major={major}&minor={minor}&sort=patch+desc',\
            { 'container_id': version['containerId'], 'major': version['major'], 'minor': version['minor'] }))['payload']['data']
        elif self.version_bump == 'minor':
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions?major={major}&sort=minor+desc',\
            { 'container_id': version['containerId'], 'major': version['major'] }))['payload']['data']
        else:
          versions = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions?sort=major+desc',\
            { 'container_id': version['containerId'] }))['payload']['data']

        if versions is not None and len(versions) > 0:
          latest_version = versions[0]

        version[self.version_bump] = latest_version[self.version_bump] + 1
        if self.version_bump in ['major', 'minor']:
          version['patch'] = 0
        if self.version_bump == 'major':
          version['minor'] = 0

        return version

  def create(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    if self.docker_image:
      container_file_path = self.docker_image_path(self.docker_image)
    else:
      container_file_path = self.container_file_path

    if not container_file_path and not self.from_version:
      image_name = input("Name of Docker Image: ")
      if not image_name:
        self.raise_exc("If you don't create from an old version then you must provide --container-file-path or --docker-image")

      container_file_path = self.docker_image_path(image_name)

    version = self.create_version_data(container_id)
    # Get version data

    if container_file_path:
      self.echo("Uploading image to Lumavate: ")
      self.echo('Image Size: {}'.format(self.util.get_size(container_file_path)))

      version['ephemeralKey'] = self.util.upload_ephemeral(container_file_path, 'application/gz')

    version['platformVersion'] = 'v2'
    if 'env' not in version.keys():
      version['env'] = {}

    version['env'] = {**version['env'], **self.resolve_env(version)}

    if self.label:
      version['label'] = self.label

    if self.port:
      version['port'] = self.port

    if self.is_editable:
      version['isEditable'] = True
      version['label'] = 'dev'
      version['editorPort'] = 5001

    if self.editor_port:
      version['editorPort'] = self.editor_port

    version['instanceCount'] = 1

    resp = self.upload_version_json(version, container_id)

    if self.json:
      if self.docker_image:
        try:
          os.remove(container_file_path)
        except:
          self.echo('Failed to delete zipped docker image...')

      click.echo(resp)
      return

    handlers = {
      'actualState': lambda x: click.style(x, fg='red') if x not in ['running'] else click.style(x, fg='green')
      }

    self.util.print_table([resp['payload']['data']], self.format, handlers=handlers)

    if self.docker_image:
      try:
        os.remove(container_file_path)
      except:
        self.echo('Failed to delete zipped docker image...')

  def upload_version_json(self, version_data, container_id):
    resp = self.util.cli_request('POST', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions',\
        {'container_id': container_id}), data=json.dumps(version_data))

    if self.json:
      return json.dumps(resp)

    # TO DO: Add progress bar after create for loading/validating container
    status_id = resp.get('payload', {}).get('data', {}).get('results', {}).get('statusId', {})
    if not status_id:
      return resp

    status_resp = self.util.show_progress(status_id, label='Uploading Container')

    error_msg = status_resp.get('payload', {}).get('data', {}).get('errorMessage')
    if error_msg is not None:
      self.raise_exc('Upload failed. {}'.format(error_msg))

    return resp

  def force_update(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    version = self.util.get_container_version(container_id, self.version)
    version_id = version.get('id')

    req_data = {'force': True}
    resp = self.util.cli_request('PUT',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}',
        {'container_id': container_id, 'version_id': version_id}), json=req_data)

    click.echo(resp)

  def download_app_zip(self):
    home = str(Path.home())
    if self.path is None:
      path = home
    else:
      path = f'{home}/{self.path}'

    for char in str(path):
      if char.isspace():
        raise click.ClickException(click.style(f"Your path cannont have a space in it - {path}", fg='red'))

    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    try:
      container = self.util.cli_request('GET',
          self.util.build_url('{experienceCloudUri}/iot/v1/containers?id={id}', {'id': container_id} ))['payload']['data']
    except:
      raise click.ClickException('Container not found')

    url_ref = container[0].get('urlRef')
    version = self.util.get_container_version(container_id, self.version)
    version_id = version.get('id')

    resp = self.util.cli_request('POST',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/token',
          {'container_id': container_id, 'version_id': version_id}), json={'type': 'editor'})

    token = resp.get('payload', {}).get('data', {}).get('token')
    service_url = self.util.ctx.get('experienceCloudUri')

    # TODO: Dynamically get integration cloud
    editor_url = f'{service_url}/ic/{url_ref}/luma-editor/download/application.zip'

    time = datetime.now().microsecond

    if platform.system() == 'Windows':
      zip_path = f'{path}\\application.{time}.zip'
    else:
      zip_path = f'{path}/application.{time}.zip'

    command_list = f"curl -L -f --create-dirs --output {zip_path} {editor_url} -H ".split()
    command_list.append(f'Authorization: Bearer {token}')
    if verify_tls is False:
      command_list.extend(['--insecure', '--proxy-insecure'])

    subprocess.run(command_list)
    click.echo(f'File Location: {zip_path}')


  def editor_info(self):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    version = self.util.get_container_version(container_id, self.version)
    version_id = version.get('id')
    url_ref = container.get('urlRef')
    editorPort = version.get('editorPort')

    if not editorPort:
      return self.old_logs(container_id, version_id, self.tail_number)

    resp = self.util.cli_request('POST',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/token',
          {'container_id': container_id, 'version_id': version_id}), json={'type': 'editor'})

    token = resp.get('payload', {}).get('data', {}).get('token')
    service_url = self.util.ctx.get('experienceCloudUri')

    headers = {'Authorization': f'Bearer {token}'}
    editor_url = f'{service_url}/ic/{url_ref}/luma-editor/version'

    req = requests.get(editor_url, headers=headers)
    click.echo(req.text)

  def tail_logs(self, editor):
    container = self.util.get_container(self.container, self.container_type)
    container_id = container.get('id')

    version = self.util.get_container_version(container_id, self.version)
    version_id = version.get('id')
    url_ref = container.get('urlRef')
    editorPort = version.get('editorPort')

    if not editorPort:
      return self.old_logs(container_id, version_id, self.tail_number)

    resp = self.util.cli_request('POST',
        self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/token',
          {'container_id': container_id, 'version_id': version_id}), json={'type': 'editor'})

    token = resp.get('payload', {}).get('data', {}).get('token')
    service_url = self.util.ctx.get('experienceCloudUri')

    if not self.tail_number:
      self.tail_number = 100

    # TODO: Dynamically get integration cloud
    editor_url = f'{service_url}/ic/{url_ref}/luma-editor/logs?tail={self.tail_number}'

    if editor:
      editor_url = f'{service_url}/ic/{url_ref}/luma-editor/logs?tail={self.tail_number}&editor=True'

    command_list = f'curl {editor_url} -H '.split()
    command_list.append(f'Authorization: Bearer {token}')
    if verify_tls is False:
      command_list.extend(['--insecure', '--proxy-insecure'])

    subprocess.run(command_list)

  def old_logs(self, container_id, version_id, tail_number):
    resp = self.util.cli_request('GET', self.util.build_url('{experienceCloudUri}/iot/v1/containers/{container_id}/versions/{version_id}/logs',\
      {'container_id': container_id, 'version_id': version_id}))

    logs = resp['payload']['data']
    if tail_number:
      tail = int(f'-{tail_number}')
      for x in logs[tail:]:
        click.echo(x)
    else:
      for x in logs:
        click.echo(x)
